"""
Copyright (C) 2016, 2017, 2020 biqqles.

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.

Functions for working with Freelancer's system layouts and navmaps.
"""
from typing import Any, Dict, Hashable, List
from collections import namedtuple
import math

from . import routines
from . import cached

PosVector = namedtuple('pos', 'x y z')
RotVector = namedtuple('rot', 'x y z')


def pos_to_sector(pos: PosVector, navmap_scale: float, divider='-', subdivider='/') -> str:
    """Convert a position vector (e.g. (-45000, 0, 75000)) into a navmap sector coordinate (e.g. 'D-5')."""
    sector_size = NAVMAP_SECTOR_SIZE / navmap_scale  # calculate size of each square
    system_size = sector_size * 8  # maximum possible x & z

    def quantise(dimension, labels):
        """Quantise a dimension `dimension` on an axis with labels `labels`."""
        magnitude = (dimension + system_size / 2) / sector_size  # "absolute magnitude" - a decimal between 0 and 7
        sector = math.floor(magnitude)
        subsector = magnitude - sector  # magnitude in the square the point rests in
        try:
            result = [labels[sector]]
        except IndexError:
            return 'Unknown'
        if subsector <= 0.2 and sector > 0:  # if it is close to the left/bottom of the square...
            result.append(labels[sector - 1])  # ...create something like B/C or 2/3
        elif subsector >= 0.8 and sector < 7:
            result.append(labels[sector + 1])
        return subdivider.join(result)

    # todo: maybe do something with pos.y < 500 >
    return divider.join(map(quantise, (pos.x, pos.z), (NAVMAP_X_LABELS, NAVMAP_Z_LABELS)))


def inter_system_route(from_system: 'System', to_system: 'System'):
    """Find the shortest route (in terms of the number of systems) between two systems."""
    return dijkstra(generate_universe_graph(), from_system, to_system)


def intra_system_route(from_solar: 'Solar', to_solar: 'Solar'):
    """Find the shortest route between two solars in the same system."""
    assert from_solar.system() == to_solar.system()
    raise NotImplementedError


@cached
def generate_universe_graph() -> Dict['System', Dict['System', int]]:
    """Generate a graph of all systems and their connections.
    Todo: calculate weight between system entrances and exits rather than using a simple placeholder."""
    return {s: {d: 1 for d in s.connections().values()} for s in routines.get_systems()}


def dijkstra(graph: Dict[Any, Dict[Any, int]], start: Hashable, end: Hashable) -> List[Hashable]:
    """An implementation of Dijkstra's algorithm using only builtin types.

     Returns the shortest path between `start` and `end` as a list of nodes in order. If no path exists, an empty list
     will be returned.

    `graph` is assumed to be a dictionary of the form {node: {connected_node: edge_weight}}."""
    distances = {node: math.inf for node in graph}
    predecessors = {}  # nodes that lie on a possible path

    distances[start] = 0  # distance from the start node is 0
    to_check = set(graph)  # the set of nodes that need to be checked

    while to_check:  # While there are still nodes to check...
        closest = min(to_check, key=distances.get)  # find the closest node to the current node
        to_check.remove(closest)

        for neighbour, weight in graph[closest].items():  # examine neighbours of closest node
            if neighbour not in to_check:
                continue
            new_distance = distances[closest] + weight
            if new_distance < distances[neighbour]:
                distances[neighbour] = new_distance
                predecessors[neighbour] = closest

    # now look through the predecessors to find the path
    path = [end]
    while path[-1] is not start:
        try:
            path.append(predecessors[path[-1]])
        except KeyError:  # no path exists
            return []
    return list(reversed(path))


NAVMAP_X_LABELS = tuple(chr(x) for x in range(ord('A'), ord('H') + 1))
NAVMAP_Z_LABELS = tuple(str(x) for x in range(1, 9))
NAVMAP_SECTOR_SIZE = 34000
DEFAULT_CRUISE_SPEED = 300
DEFAULT_LANE_SPEED = 1000



"""
Copyright (C) 2016, 2017, 2020 biqqles.

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
from types import FunctionType
from typing import Dict, Any
from functools import wraps
import warnings

warnings.formatwarning = lambda message, *args, **kwargs: f'{message!s}\n'  # patch formatter to only show message
central_cache: Dict[FunctionType, Dict[tuple, Any]] = {}


def cached(function):
    """A decorator which caches a function to the central cache."""
    central_cache[function] = {}  # initialise cache

    @wraps(function)
    def wrapped(*args):
        function_cache = central_cache[function]
        if args in function_cache:
            return function_cache[args]
        result = function_cache[args] = function(*args)
        return result
    return wrapped


from . import paths
from .routines import get_commodities, get_bases, get_equipment, get_ships, get_systems, get_factions, get_goods


shorthand = {'bases': get_bases,
             'commodities': get_commodities,
             'equipment': get_equipment,
             'factions': get_factions,
             'goods': get_goods,
             'ships': get_ships,
             'systems': get_systems}


# shorthand for Python 3.7 and up
def __getattr__(name):
    if not paths.install:
        raise AssertionError('No path set')
    if name in shorthand:
        return shorthand[name]()
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



"""
Copyright (C) 2016, 2017, 2020 biqqles.

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.

This module provides utilities for working with Freelancer's paths.
"""
import os
from typing import Dict, Tuple

from .formats import ini
from . import cached

install: str
inis: Dict[str, Tuple[str]] = {}  # ini category (defined in freelancer.ini) to a list of paths
dlls: Dict[int, str] = {}  # dll number to path


def set_install_path(new_path, discovery=False):
    """Lists the path to the installation"""
    global install
    if not os.path.exists(new_path):
        raise FileNotFoundError(new_path)
    assert is_probably_freelancer(new_path, discovery)
    install = new_path
    generate_index()


@cached
def fix_path_case(path: str) -> str:
    """Thanks to the nature of Windows' file systems, Freelancer frequently uses apparently arbitrary casing for many
    paths. This function takes an absolute path and returns it with the casing as it really is on the filesystem.

    This function mixes iteration/recursion to make the use of dynamic programming more effective. If one recursive
    function with an additional tail accumulator was used, it would prevent the cache from being effective at storing
    precomputed known-correct paths."""

    @cached
    def correct_subpath(path_: str) -> str:
        """Recursively reduce a path from its end until it diverges from the filesystem's structure. Then, find the
        correct case for that divergent level and return the subpath."""
        head, tail = os.path.split(path_)
        if os.path.exists(head):  # go back till divergence
            #  tail is now divergent level; find its correct case (taking the first match)
            correct_case = next(filter(lambda s: s.casefold() == tail.casefold(), os.listdir(head)), None)
            if correct_case is None:  # if no match possible
                raise FileNotFoundError
            return os.path.join(head, correct_case)
        return correct_subpath(head)

    path = path.rstrip(r'\/')  # remove trailing slashes which mess up path.split
    if os.name == 'nt':
        return path  # on Windows path.exists ignores case, so no point in continuing

    result = correct_subpath(path)
    while len(result) != len(path):
        result = correct_subpath(result + path[len(result):])
    return result


def is_probably_freelancer(path, discovery=False):
    """Verifies that the given directory is (probably) an actual Freelancer installation, by checking that identifying
    files and directories are present."""
    identifiers = {'DATA', 'DLLS', 'EXE'}
    if discovery:
        identifiers |= {'DSLauncher.exe'}
    return os.path.isdir(path) and identifiers.issubset(os.listdir(path))


def construct_path(*subpath):
    """Form an absolute path to a file in the install directory based on a subpath relative to its root."""
    candidate = os.path.join(install, *subpath).replace('\\', '/')
    try:
        return fix_path_case(candidate)
    except FileNotFoundError:  # todo: hacky, should not handle invalid path silently
        return candidate


def generate_index():
    """Use freelancer.ini to build an index of inis and dlls."""
    freelancer_ini = os.path.join(install, 'EXE/freelancer.ini')
    parsed = ini.sections(freelancer_ini, fold_values=False)

    root = parsed['freelancer'][0]  # todo: also look at 'data path'
    resources = parsed['resources'][0]['dll']
    data = parsed['data'][0]
    data.update(root)

    dlls.update({i: construct_path('EXE', f) for i, f in enumerate(resources, 1)})  # dll 0 is Freelancer.exe itself
    inis.update({category: tuple(construct_path('DATA', f) for f in files) for category, files in data.items()})



"""
Copyright (C) 2016, 2017, 2020 biqqles.

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.

This file implements a reader for Freelancer's "BINI" (binary INI)
format. Thanks to Bas Westerbaan for providing excellent
documentation available here: <http://blog.w-nz.com/uploads/bini.pdf>.
"""
from typing import Dict, List
from collections import defaultdict
from struct import unpack
from os.path import getsize, isfile

VALUE_TYPES = {1: 'i', 2: 'f', 3: 'i'}  # maps a byte value type to a struct format string


def parse_file(path: str, fold_values=True, lower=True):
    """Read the BINI file at `path` and produce an output of the form
    {section_name -> [{entry_name -> entry_values}]}"""
    result = []
    string_table = {}
    file_size = getsize(path)

    with open(path, 'rb') as f:
        # read file header
        magic, version, str_table_offset = unpack('4sII', f.read(12))
        assert magic == b'BINI', version == 1

        # read string table, which stretches from str_table_offset to EOF
        f.seek(str_table_offset)
        raw_table = f.read(file_size - str_table_offset - 1)

        count = 0
        for s in raw_table.split(b'\0'):
            string_table[count] = s.decode('cp1252').lower()
            count += len(s) + 1

        # return to end of header to read sections
        f.seek(12)

        while f.tell() < str_table_offset:
            # read section header
            section_name_ptr, entry_count = unpack('hh', f.read(4))
            section_name = string_table[section_name_ptr]

            section_entries = []
            for e in range(entry_count):
                # read entry
                entry_name_ptr, value_count = unpack('hb', f.read(3))
                entry_name = string_table[entry_name_ptr]
                entry_values = []

                for v in range(value_count):
                    # read value
                    value_type, = unpack('b', f.read(1))
                    value_data, = unpack(VALUE_TYPES[value_type], f.read(4))

                    if value_type == 3:
                        # it is a pointer relative to the string table
                        value_data = string_table[value_data]

                    entry_values.append(value_data)

                if value_count > 1:
                    entry_value = tuple(entry_values)
                elif value_count == 1:
                    entry_value = entry_values[0]
                else:
                    continue

                section_entries.append((entry_name, entry_value))
            result.append((section_name, section_entries))

    return tuple(result)


def dump(path: str) -> str:
    """Dump the BINI file at `path` to an INI-formatted string."""
    bini = parse_file(path, fold_values=False)

    lines = []
    for section_name, sections in bini:
        for entry in sections:
            lines.append(f'[{section_name}]')
            # convert the entries in this section to strings and add to output
            for key, values in entry:
                # form key value pairs for each entry value. Expand tuples to remove quotes and brackets
                entries = [f'{key} = {", ".join(map(str, v)) if isinstance(v, tuple) else v}' for v in values]
                lines.extend(entries)
            lines.append('')  # add a blank line after each section
    return '\n'.join(lines)


def is_bini(path: str) -> bool:
    """Returns whether the (.ini) file at `path` is a BINI by checking its magic number."""
    assert isfile(path)
    with open(path, 'rb') as f:
        data = f.read(4)
    return data[:4] == b'BINI'



"""
Copyright (C) 2016, 2017, 2020 biqqles.

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.

This file defines entities for equipment. 'Equipment' includes anything
which can be mounted on a ship or carried in its hold.

There are also several non-entity section types that appear in equipment
files. These include Lod, TradeLane, InternalFX, AttachedFX, Explosion,
Light, Motor, LootCrate, and Shield. Currently these are excluded as
they do not exactly fit flint's entity model, and partly for the sake
of simplicity in the first incarnation of equipment parsing.

Because of the sheer number of classes in this file, apart from abstract
classes they currently have a 1-1 mapping with INI section types to
simplify and optimise instantiation logic. So, for example, subtypes of
Gun could be Missile and Turret, but instead you should determine if a
Gun instance is one of these by checking its attributes.
"""
from typing import Dict, Optional, Tuple, cast
import math

from . import Entity
from .. import routines


class Equipment(Entity):
    """Abstract class for something which can be mounted on a ship or carried in its hold."""
    lootable: bool = False

    def icon(self) -> bytes:
        """This equipment's icon in TGA format."""
        return self.good().icon()

    def good(self) -> Optional['Good']:
        """The good entity for this piece of equipment."""
        return routines.get_goods().of_type(EquipmentGood).unique(equipment=self.nickname)

    def sold_at(self) -> Dict['Base', int]:
        """A dict of bases that sell this good of the form {base: price}. All bases buy equipment."""
        return self.good().sold_at() if self.good() else {}

    def price(self) -> int:
        """The default price of this equipment."""
        return self.good().price if self.good() else 0

    def is_valid(self) -> bool:
        """Whether the equipment is valid, i.e. it defines a good. Note that some specialised equipment subtypes
        do not have goods, rendering this method meaningless for them."""
        return self.good() is not None


class Mountable(Equipment):
    """Abstract class for a piece of equipment that is mountable."""
    volume: int = 0  # volume of one unit in ship's cargo bay


class External(Mountable):
    """Abstract class for a piece of mountable equipment that is externally mounted, and therefore typically
    destructible."""
    hit_pts: int


# equipment typically defined in weapon_equip.ini
class Weapon(External):
    """Abstract class for a piece of external equipment that is a weapon."""
    refire_delay: float
    projectile_archetype: str

    def refire(self) -> float:
        """The refire value as displayed in-game, which is the reciprocal of refire_delay."""
        return 1 / self.refire_delay

    def hull_damage(self) -> float:
        """Hull damage dealt per shot."""
        raise NotImplementedError

    def shield_damage(self) -> float:
        """Shield damage dealt per shot."""
        raise NotImplementedError

    def projectile(self) -> 'Projectile':
        """The Projectile fired or dropped by this weapon."""
        return routines.get_equipment().get(self.projectile_archetype)


class Gun(Weapon):
    """A gun that goes 'pew'. Not much to be said."""
    power_usage: float
    muzzle_velocity: int
    projectile_archetype: str
    hp_gun_type: Optional[str] = None  # NPC guns lack this field
    dispersion_angle: float = 0.0
    dry_fire_sound: Optional[str] = None  # only missiles (and mine droppers, but they're a different class) have this
    auto_turret: bool  # only turrets have this

    def munition(self) -> Optional['Munition']:
        """The Munition fired by this weapon."""
        return cast(Munition, self.projectile())

    def hull_damage(self) -> float:
        """Hull damage dealt per shot."""
        return self.munition().hull_damage_()

    def shield_damage(self) -> float:
        """Shield damage dealt per shot."""
        return self.munition().energy_damage_()

    def hull_dps(self) -> float:
        """Hull damage dealt per second."""
        return self.refire() * self.munition().hull_damage

    def shield_dps(self) -> float:
        """Shield damage dealt per second."""
        return self.refire() * self.munition().energy_damage

    def energy_per_second(self) -> float:
        """Energy consumption per second (i.e. power).."""
        return self.refire() * self.power_usage

    def efficiency(self) -> float:
        """Energy consumption per second (i.e. power).."""
        return ((self.hull_damage() + self.shield_damage()) / self.power_usage) if self.power_usage else 0.0

    def technology(self) -> Optional[str]:
        """The technology of this gun. Null for non-energy (e.g. kinetic) guns."""
        return self.munition().weapon_type

    def range(self) -> int:
        """The range this weapon can shoot to."""
        return int(self.muzzle_velocity * self.munition().lifetime)

    def is_valid(self) -> bool:
        """Whether the gun is valid, i.e. it defines a good and a munition."""
        return super().is_valid() and (self.munition() is not None)

    def is_missile(self) -> bool:
        """Whether the gun is a missile/torpedo launcher. Another, much slower, way to test this would be to check
        if the gun's munition has a "motor" field."""
        return bool(self.dry_fire_sound) or self.munition().cruise_disruptor is not None or 'Torpedo' in self.name()

    def is_turret(self) -> bool:
        """Whether the gun is a turret."""
        return self.auto_turret or (self.hp_gun_type and 'turret' in self.hp_gun_type) or 'Turret' in self.name()


class MineDropper(Weapon):
    """A dispenser for mines."""

    def mine(self) -> Optional['Mine']:
        """The Mine dropped by this weapon."""
        return cast(Mine, self.projectile())

    def hull_damage(self) -> float:
        """The hull damage inflicted on contact with this dropper's mine."""
        return self.mine().explosion().hull_damage

    def shield_damage(self) -> float:
        """The shield damage inflicted on contact with this dropper's mine."""
        return self.mine().explosion().energy_damage


class CloakingDevice(External):
    """A cloaking device. Used in cutscenes in the campaign as well as on servers with cloaking devices enabled through
    FLHook."""


class Projectile(Equipment):
    """Abstract. A projectile fired or dropped by a Weapon."""
    lifetime: float  # time in seconds that the projectile lingers in space before despawning


class Mine(Projectile):
    """An explosive mine that can be dropped into space."""
    explosion_arch: str  # nickname of Explosion
    seek_dist: int
    top_speed: int
    acceleration: int
    ammo_limit: int

    def explosion(self) -> Optional['Explosion']:
        """The Explosion triggered by this mine."""
        return routines.get_equipment().get(self.explosion_arch)


class Munition(Projectile):
    """A projectile fired by a Weapon."""
    hp_type: str
    hull_damage: int = 0
    energy_damage: int = 0
    requires_ammo: bool = True  # todo: not sure about this default
    weapon_type: Optional[str] = None  # present only for energy weapons
    ammo_limit: int = math.inf
    explosion_arch: Optional[str] = None  # nickname of Explosion
    cruise_disruptor: Optional[bool] = None  # only set for missiles
    motor: Optional[str] = None  # only set for missiles
    seeker: Optional[str] = None

    def explosion(self) -> Optional['Explosion']:
        """The Explosion triggered by this munition."""
        return routines.get_equipment().get(self.explosion_arch) if self.explosion_arch else None

    def hull_damage_(self) -> int:
        """The hull damage inflicted by this munition, taking into consideration its explosion."""
        try:
            return self.hull_damage or self.explosion().hull_damage
        except AttributeError:
            return 0

    def energy_damage_(self) -> int:
        """The shield damage inflicted by this munition, taking into consideration its explosion."""
        try:
            return self.energy_damage or self.explosion().energy_damage
        except AttributeError:
            return 0

    def motor_(self) -> Optional['Motor']:
        """This Munition's Motor."""
        return routines.get_equipment().get(self.motor) if self.motor else None


class Motor(Projectile):
    """A missile motor."""
    accel: int
    delay: int


class Explosion(Projectile):
    """A very strange thing to call equipment, whatever way you look at it, yet defined in weapon_equip."""
    lifetime: Tuple[int, int]
    radius: int
    hull_damage: int
    energy_damage: int
    strength: int


# equipment typically defined in st_equip.ini
class Thruster(External):
    """A thruster that provides supplementary acceleration and velocity to the main engine."""
    power_usage: float


class ShieldGenerator(External):
    """A piece of equipment that generates a shield bubble around a ship which absorbs damage from weapon fire."""
    shield_type: str = ''  # the shield's technology type
    max_capacity: float
    explosion_resistance: float = 0.0


# equipment typically defined in misc_equip.ini
class Power(Mountable):
    """A ship's power plant."""
    capacity: int
    charge_rate: int


class Tractor(Mountable):
    """A tractor beam generator."""
    max_length: int  # range of beam in M


class Scanner(Mountable):
    """A scanner, akin to a radar transmitter/receiver."""
    range: int  # maximum contact acquisition range in M
    cargo_scan_range: int  # maximum cargo scan range in M


class CounterMeasure(Equipment):
    """A countermeasure that can be deployed against seeking missiles."""
    lifetime: float
    range: int
    diversion_pctg: float
    ammo_limit: int = math.inf  # typically only NPC CMs have unlimited ammo

    def effectiveness(self) -> float:
        """The probability this countermeasure will defeat an incoming missile."""
        return self.diversion_pctg / 100


class CounterMeasureDropper(Weapon):
    """A countermeasure dispenser."""

    def countermeasure(self) -> Optional['CounterMeasure']:
        """The CounterMeasure launched by this dropper."""
        return routines.get_equipment().get(self.projectile_archetype)


class RepairKit(Equipment):
    """A nanobot that can be used to repair hull damage."""


class ShieldBattery(Equipment):
    """A shield battery, used to recharge shields instantaneously."""


# equipment typically defined in engine_equip.ini
class Engine(Mountable):
    """A reaction engine that must be mounted to a ship to provide propulsion."""
    cruise_charge_time: int
    reverse_fraction: float
    linear_drag: float
    max_force: float
    

# equipment typically defined in select_equip.ini
class Armor(Mountable):
    """An armour upgrade."""
    hit_pts_scale: float  # multiplier applied to ship hull hit points when mounted


class CargoPod(Mountable):
    """A cargo pod. These can absorb damage but this is their only function in the final game."""


class Commodity(Equipment):
    """A Commodity is a tradeable piece of "equipment". Unlike other forms of equipment, commodities can typically
    be bought and sold for variable amounts on different bases."""
    decay_per_second: int
    volume: int  # volume of one unit in ship's cargo bay

    def price_at(self):
        pass

    def highest_price(self):
        return

    def bought_at(self) -> Dict['Base', int]:
        """A dict of bases that buy this commodity of the form {base: price}."""
        return self.good().bought_at()


from .goods import Good, EquipmentGood



"""
Copyright (C) 2016, 2017, 2020 biqqles.

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
from typing import Tuple, List, Optional
import math

from . import Entity, EntitySet
from .goods import ShipHull, ShipPackage
from .equipment import Equipment, Power, Engine
from ..formats import dll
from .. import routines


LOG_OF_E = math.log10(math.e)  # used to approximate angular acceleration curve


class Ship(Entity):
    """A starship with a cargo bay and possibly hardpoints for weapons."""
    ids_info1: int  # ship description (ids_info stores statistics)
    ids_info2: int  # extra stat names list
    ids_info3: int  # extra stat values list
    ship_class: int
    hit_pts: int
    hold_size: int
    mass: int
    linear_drag: int
    nanobot_limit: int = 0
    shield_battery_limit: int = 0
    steering_torque: Tuple[float, float, float]
    angular_drag: Tuple[float, float, float]
    rotation_inertia: Tuple[float, float, float]

    def hull(self) -> Optional[ShipHull]:
        """This ship's hull entity."""
        return routines.get_goods().of_type(ShipHull).unique(ship=self.nickname)

    def package(self) -> Optional[ShipPackage]:
        """This ship's package entity."""
        if not self.hull():
            return None
        return routines.get_goods().of_type(ShipPackage).unique(hull=self.hull().nickname)

    def sold_at(self) -> List['Base']:
        """A list of bases which sell this ship."""
        if not self.package():
            return []
        return list(self.package().sold_at().keys())

    def price(self) -> int:
        """The price of this ship's hull."""
        try:
            return self.hull().price
        except AttributeError:
            return 0

    def icon(self) -> bytes:
        """This ship's icon."""
        return self.hull().icon()

    def infocard(self, markup='html') -> str:
        """I have no idea why the order these are displayed in is not ascending, but anyway."""
        if markup == 'html':
            lookup = dll.lookup_as_html
        elif markup == 'plain':
            lookup = dll.lookup_as_plain
        elif markup == 'rdl':
            lookup = dll.lookup
        else:
            raise ValueError
        return '<p>'.join(map(lookup, (self.ids_info1, self.ids_info)))

    def type(self) -> Optional[str]:
        """The name of the type (class) of this ship."""
        return self.TYPE_ID_TO_NAME.get(self.ship_class)

    def turn_rate(self) -> float:
        """The maximum turn rate (i.e. angular speed) of this ship, in rad/s."""
        return self.steering_torque[0] / self.angular_drag[0]

    def angular_distance_in_time(self, time=1) -> float:
        """The angular displacement in radians from rest to `time` (in seconds).
        TODO: Currently only an approximation."""
        return self.turn_rate() / self.response()

    def response(self) -> float:
        """The "response time" is defined as the time to reach 90% maximum angular speed (in seconds).
        Thanks to Haste for this formula."""
        return self.rotation_inertia[0] / (self.angular_drag[0] * LOG_OF_E)

    def equipment(self) -> EntitySet[Equipment]:
        """The set of this ship package's equipment upon purchase."""
        package = self.package()
        return package.equipment() if package else EntitySet([])

    def power_core(self) -> Optional[Power]:
        """The Power entity for this ship."""
        return self.equipment().of_type(Power).first

    def engine(self) -> Optional[Engine]:
        """The Power entity for this ship."""
        return self.equipment().of_type(Engine).first

    def impulse_speed(self) -> float:
        """The maximum forward impulse (non-cruise) speed of this ship (m/s)."""
        engine = self.engine()
        return self.linear_speed(engine.max_force) if engine else 0

    def reverse_speed(self) -> float:
        """The maximum reverse speed of this ship (m/s)."""
        engine = self.engine()
        return self.impulse_speed() * engine.reverse_fraction if engine else 0

    def linear_speed(self, force: float) -> float:
        """The maximum speed of this ship for a given force, in m/s."""
        return force / self.total_linear_drag()

    def total_linear_drag(self) -> float:
        """The total linear resistive force for this ship and engine, in N (?)."""
        engine = self.engine()
        try:
            return engine.linear_drag + self.linear_drag if engine else self.linear_drag
        except TypeError:
            return 1

    def cruise_charge_time(self):
        """The time taken to charge this ship's cruise engine, in seconds."""
        engine = self.engine()
        return engine.cruise_charge_time if engine else 0

    TYPE_ID_TO_NAME = {0: 'Light Fighter',
                       1: 'Heavy Fighter',
                       2: 'Freighter',
                       # Discovery:
                       3: 'Very Heavy Fighter',
                       4: 'Super Heavy Fighter',
                       5: 'Bomber',
                       6: 'Transport',  # more specifically;
                       7: 'Transport',  # trains
                       8: 'Transport',  # battle-transports
                       9: 'Transport',
                       10: 'Transport',  # liners.
                       11: 'Gunboat',  # gunships
                       12: 'Gunboat',
                       13: 'Cruiser',  # destroyers
                       14: 'Cruiser',
                       15: 'Cruiser',  # battlecruisers
                       16: 'Battleship',
                       17: 'Battleship',  # carriers
                       18: 'Battleship',  # flagships
                       19: 'Freighter'}  # repair ships



"""
Copyright (C) 2016, 2017, 2020 biqqles.

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.

This file provides a read-only implementation of Universal Tree Format
(UTF), a hierarchical binary format developed by Digital Anvil. In
Freelancer it is used to store binary assets, like textures and icons.

Reference: <https://wiki.librelancer.net/utf:universal_tree_format>
"""
import deconstruct as c
from . import WinStruct


def parse(path):
    result = {}
    with open(path, 'rb') as f:
        header = UtfHeader(f.read(56))

        entry_count = header.TreeSize // header.EntrySize

        # read name dictionary
        f.seek(header.NamesOffset)
        dictionary = f.read(header.NamesUsedSize).split(b'\0')
        position = 0

        names = {}
        for name in dictionary:
            names[position] = name.decode('ascii')
            position += len(name) + 1

        # read data tree
        f.seek(header.TreeOffset)
        for e in range(entry_count):
            entry = UtfEntry(f.read(44))
            position = f.tell()
            name = names[entry.DictionaryNameOffset]
            f.seek(entry.ChildOrDataOffset + header.DataStartOffset)
            yield name, f.read(entry.UsedDataSize)
            f.seek(position)
        return result


def extract(path, target_directory):
    for dir_name, dir_data in parse(path):
        if dir_name == target_directory:
            return dir_data
    raise KeyError


def dump(path):
    return dict(parse(path))


TYPE_CHILD = 0x80  # utf tree types
TYPE_DATA = 0x10


class UtfHeader(WinStruct):
    Signature: c.uint32
    Version: c.uint32
    TreeOffset: c.uint32
    TreeSize: c.uint32
    _0: c.uint32
    EntrySize: c.uint32
    NamesOffset: c.uint32
    NamesAllocatedSize: c.uint32
    NamesUsedSize: c.uint32
    DataStartOffset: c.uint32
    _1: c.uint32
    _2: c.uint32
    FiletimeLow: c.uint32
    FiletimeHigh: c.uint32


class UtfEntry(WinStruct):
    NextSiblingOffset: c.uint32
    DictionaryNameOffset: c.uint32
    EntryType: c.uint32
    SharingAttributes: c.uint32
    ChildOrDataOffset: c.uint32
    AllocatedDataSize: c.uint32
    UsedDataSize: c.uint32
    UncompressedDataSize: c.uint32
    CreationTime: c.uint32
    LastAccessTime: c.uint32
    LastWriteTime: c.uint32



"""
Copyright (C) 2016, 2017, 2020 biqqles.

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.

This file implements a platform-independent resource DLL reader, as
well as the logic that Freelancer uses to map those resources to
internal IDs. Additionally it implements conversion from RDL (used
for rich-text strings) to HTML.
"""
from typing import Dict
from os import SEEK_CUR

import deconstruct as c

from ..interface import rdl_to_html, rdl_to_plaintext
from .. import cached
from .. import paths
from . import WinStruct


resource_table: Dict[int, Dict[int, str]] = {}


@cached
def lookup(resource_id: int) -> str:
    """Looks up the text associated with a resource ID (string or HTML) in the resource dlls."""
    if resource_id is None:  # sometimes objects which should have infocards don't. Freelancer doesn't seem to care
        return ''

    # all references to string/html resources in the inis (strid or ids_info) use "external ids" which are based off the
    # positions of the dlls in Freelancer.ini/[Resources]
    dll_no = resource_id // 65536  # get the dll index the id refers to
    external_id_offset = dll_no * 65536  # the external id of position 0 in the dll

    if dll_no not in paths.dlls:  # likewise
        return ''

    # if dll has already been loaded
    if dll_no in resource_table:
        return resource_table[dll_no].get(resource_id, '')  # a resource id that maps to nothing is also an empty string

    # otherwise, load in the new dll, and call again
    dll_path = paths.dlls[dll_no]
    resource_table[dll_no] = parse(dll_path, external_id_offset)
    return lookup(resource_id)


@cached
def lookup_as_html(resource_id: int) -> str:
    """Looks up the given resource ID and translates RDL to HTML."""
    return rdl_to_html(lookup(resource_id))


@cached
def lookup_as_plain(resource_id: int) -> str:
    """Looks up the given resource ID and strips out all RDL tags. Paragraph tags are replaced with newlines."""
    return rdl_to_plaintext(lookup(resource_id))


def parse(path: str, external_strid_offset: int) -> Dict[int, str]:
    """Read the DLL at the given path into a mapping of external ids to string/XML resources.
    Format reference: <https://docs.microsoft.com/en-gb/windows/win32/debug/pe-format>"""

    def read_rdt(entry_count: int):
        """Read the entries in a resource directory table starting at the file cursor."""
        rdt_entries = {}
        for e in range(entry_count):
            entry = ResourceDirectoryEntry(f.read(8))
            rdt_entries[entry.IntegerID] = entry.Offset & 0x7FFFFFFF  # only interested in lower 31 bits
        return rdt_entries

    def read_string_table():
        """Read a string table starting at the file cursor."""
        string_table = {}
        for s in range(16):  # sixteen strings are allocated per table
            resource_string = ResourceDirectoryString(f.read(2))
            if resource_string.Length:
                strid = (name - 1) * 16 + s + external_strid_offset
                text = f.read(resource_string.Length * 2).decode('utf-16')
                string_table[strid] = text
        return string_table

    with open(path, 'rb') as f:
        # read PE signature
        f.seek(0x3C)  # find offset to signature
        pe_signature_offset = ord(f.read(1))
        f.seek(pe_signature_offset)
        assert f.read(4) == b'PE\0\0'

        # read COFF header, which begins immediately after PE signature
        coff = CoffHeader(f.read(20))
        # skip the optional header (is irrelevant to us) to get to the section headers
        f.seek(coff.SizeOfOptionalHeader, SEEK_CUR)

        # go through section headers to find .rsrc
        for i in range(coff.NumberOfSections):
            section = SectionHeader(f.read(40))
            if section.Name.rstrip(b'\0') == b'.rsrc':
                rsrc_offset = section.PointerToRawData
                break
        else:
            raise EOFError('.rsrc section not found')

        # read Resource Directory Table header for .rsrc section
        f.seek(rsrc_offset)
        resource_directory_table = ResourceDirectoryTable(f.read(16))

        # remember that named entries precede ids

        # read type directory entries. For types, only ids are used, not string names
        resource_types = read_rdt(resource_directory_table.IdEntryCount)  # resource types to offsets

        # for each resource type, read its name table
        name_offsets = {}
        for resource_type, name_table_offset in resource_types.items():  # for each data type
            f.seek(name_table_offset + rsrc_offset)
            name_entries = ResourceDirectoryTable(f.read(16))
            name_offsets[resource_type] = read_rdt(name_entries.IdEntryCount)

        resources = {}

        for resource_type, name_offset in name_offsets.items():
            for name, description_offset in name_offset.items():
                f.seek(description_offset + rsrc_offset)
                # commence reading another section header - I actually have no idea why this is here but it matches
                data_section = SectionHeader(f.read(40))
                f.seek(rsrc_offset + data_section.PointerToRawData)  # jump there

                # read Resource Data Entry
                data = ResourceDataEntry(f.read(16))  # could also get codepage int here ?
                f.seek(data.DataRVA)

                if resource_type == RT_STRING:
                    resources.update(read_string_table())
                elif resource_type == RT_HTML:
                    strid = name + external_strid_offset
                    text = f.read(data.Size).decode('utf-16')
                    resources[strid] = text
                elif resource_type == RT_VERSION:
                    pass
                else:
                    raise NotImplementedError('Unexpected resource type:', hex(resource_type))
        return resources


# win32 resource types <https://docs.microsoft.com/en-us/windows/win32/menurc/resource-types>
RT_HTML = 0x17  # html resource type (in Freelancer, used for XML-encoded rich text)
RT_STRING = 0x06  # string table entry resource type
RT_VERSION = 0x10  # version information - ignored


# PE format structs
class CoffHeader(WinStruct):
    Machine: c.int16
    NumberOfSections: c.int16
    TimeDateStamp: c.int32
    PointerToSymbolTable: c.int32
    NumberOfSymbols: c.int32
    SizeOfOptionalHeader: c.int16
    Characteristics: c.int16


class SectionHeader(WinStruct):
    Name: c.char[8]
    VirtualSize: c.int32
    VirtualAddress: c.int32
    SizeOfRawData: c.int32
    PointerToRawData: c.int32
    PointerToRelocations: c.int32
    PointerToLinenumbers: c.int32
    NumberOfRelocations: c.int16
    NumberOfLinenumbers: c.int16
    Characteristics: c.int32


class ResourceDirectoryTable(WinStruct):
    Characteristics: c.int32
    TimeDateStamp: c.int32
    MajorVersion: c.int16
    MinorVersion: c.int16
    NameEntryCount: c.int16
    IdEntryCount: c.int16


class ResourceDirectoryEntry(WinStruct):
    IntegerID: c.int32
    Offset: c.int32  # if MSB is 0, this is the address of a leaf RDE. If 1, lower bits are the RDT one level down


class ResourceDataEntry(WinStruct):
    DataRVA: c.int32
    Size: c.int32
    Codepage: c.int32
    Reserved: c.int32


class ResourceDirectoryString(WinStruct):
    Length: c.int16



"""
Copyright (C) 2016, 2017, 2020 biqqles.

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
from setuptools import setup, find_packages

setup(
    name='fl-flint',  # distribution name
    version='0.8',

    author='biqqles',
    author_email='biqqles@protonmail.com',
    url='https://github.com/biqqles/flint',

    description='A parser for Freelancer and its formats',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',

    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Operating System :: OS Independent',
        'Topic :: Games/Entertainment',
        'Intended Audience :: Developers',
        'Development Status :: 4 - Beta'
    ],

    install_requires=open('requirements.txt').readlines(),
    python_requires='>=3.6',
)



"""
Copyright (C) 2016, 2017, 2020 biqqles.

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.

This file contains routines for parsing specific sets of information
from the game files. All exported functions return EntitySets.
"""
from typing import Dict, Union
from collections import defaultdict
import warnings

from . import paths
from . import cached
from .formats import ini
from .maps import PosVector

from .entities import EntitySet
from .entities import Good, EquipmentGood, CommodityGood, ShipHull, ShipPackage
from .entities import Commodity, Equipment, Armor, ShieldGenerator, Thruster, Gun, Engine, Power, ShieldBattery, \
    CounterMeasure, CounterMeasureDropper, Scanner, Tractor, CargoPod, CloakingDevice, RepairKit, Mine, MineDropper, \
    Munition, Explosion, Motor
from .entities import Ship
from .entities import Base, System, Faction
from .entities import Solar, Object, Jump, BaseSolar, Star, Planet, PlanetaryBase, TradeLaneRing, Wreck, Zone
from . import entities


@cached
def get_systems() -> EntitySet[System]:
    """All systems defined in the game files."""
    systems = ini.sections(paths.inis['universe'])['system']
    return EntitySet(System(**s, ids_name=s.pop('strid_name')) for s in systems if 'file' in s)


@cached
def get_bases() -> EntitySet[Base]:
    """All bases defined in the game files."""
    bases = ini.sections(paths.inis['universe'])['base']
    return EntitySet(Base(**b, ids_name=b['strid_name']) for b in bases if 'strid_name' in b)


@cached
def get_factions() -> EntitySet[Faction]:
    """All groups (i.e. factions) defined in the game files."""
    groups = ini.sections(paths.inis['initial_world'])['group']
    return EntitySet(Faction(**g) for g in groups)


@cached
def get_goods() -> EntitySet[Good]:
    """All goods defined in the game files."""
    goods = ini.sections(paths.inis['goods'])['good']
    result = []

    for g in goods:
        if g['category'] == 'ship':
            result.append(ShipPackage(**g))
        elif g['category'] == 'equipment':
            result.append(EquipmentGood(**g))
        elif g['category'] == 'commodity':
            result.append(CommodityGood(**g))
        elif g['category'] == 'shiphull':
            result.append(ShipHull(**g))
        else:
            result.append(Good(**g))
    return EntitySet(result)


@cached
def get_equipment() -> EntitySet[Equipment]:
    """All equipment defined in the game files."""
    section_name_to_type = {name.lower(): globals()[name] for name in dir(entities) if name in globals()}
    excluded_sections = {'light', 'tradelane', 'internalfx', 'attachedfx', 'shield', 'lod', 'lootcrate'}

    equipment = ini.parse(paths.inis['equipment'])

    def generate_entities():
        """Convert equipment sections to entities."""
        for section, contents in equipment:
            if section in excluded_sections:
                continue  # not really entities, see docstring for equipment.py
            if section in section_name_to_type:
                try:
                    yield section_name_to_type[section](**contents)
                except TypeError as e:
                    warnings.warn(f'Failed to initialise equipment of type {section!r} and nickname '
                                  f'{contents.get("nickname")!r}: {e.args[0]}')
            else:
                warnings.warn(f'Unknown equipment type {section!r} - ignoring')

    return EntitySet(generate_entities())


@cached
def get_commodities() -> EntitySet[Commodity]:
    """All commodities defined in the game files. Commodities are actually a type of equipment, so this function
    is for convenience's sake."""
    return get_equipment().of_type(Commodity)


@cached
def get_ships() -> EntitySet[Ship]:
    """All ships defined in the game files."""
    ships = ini.sections(paths.inis['ships'])['ship']
    result = []

    for s in ships:
        if 'ids_info3' in s:
            result.append(Ship(**s))

    return EntitySet(result)


@cached
def get_system_contents(system: System) -> EntitySet[Solar]:
    """All solars (objects and zones) in a given system."""
    result = []
    contents = ini.parse(system.definition_path())

    # categorise objects based on their keys
    for solar_type, attributes in contents:
        if 'ids_name' not in attributes:
            continue
        attributes['_system'] = system
        attributes['pos'] = PosVector(*attributes['pos'])

        if solar_type == 'object':
            o = attributes
            keys = o.keys()
            if {'base', 'reputation', 'space_costume'} <= keys:
                result.append(BaseSolar(**o))
            elif 'goto' in keys:
                result.append(Jump(**o))
            elif 'prev_ring' in keys or 'next_ring' in keys:
                result.append(TradeLaneRing(**o))
            elif 'loadout' in keys and 'reputation' not in keys:
                result.append(Wreck(**o))
            elif 'star' in keys:
                result.append(Star(**o))
            elif 'spin' in keys:
                result.append(PlanetaryBase(**o) if 'base' in keys else Planet(**o))
            else:
                result.append(Object(**o))

        elif solar_type == 'zone':
            z = attributes
            result.append(Zone(**z))

    return EntitySet(result)


@cached
def get_markets() -> Dict[Union[Base, Good], Dict[bool, Dict[Union[Good, Base], int]]]:
    """Market (i.e. economy) data for the universe. Result is of the form
    {base/good nickname -> {whether sold -> (good/base entity, price at base}}."""
    market = ini.sections(paths.inis['markets'], fold_values=False)['basegood']

    goods = get_goods()
    bases = get_bases()

    result = defaultdict(lambda: {True: {}, False: {}})
    for b in market:
        try:
            base = b['base'][0]
            base_entity = bases[base]
        except IndexError:
            warnings.warn('BaseGood has no base')
            continue

        for good, min_rank, min_rep, min_stock, max_stock, depreciate, multiplier, *_ in b['marketgood']:
            if not multiplier:
                continue

            try:
                good_entity = goods[good]
            except KeyError:
                warnings.warn(f'BaseGood for base {base!r} refers to undefined good: {good!r}')
                continue
            sold = not (min_stock == 0 or max_stock == 0)
            price_at_base = int(round(good_entity.price * multiplier))

            result[base_entity][sold][good_entity] = price_at_base
            result[good_entity][sold][base_entity] = price_at_base
    return result



"""
Copyright (C) 2016, 2017, 2020 biqqles.

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.

This namespace contains definitions for entities found within
Freelancer.
"""
from typing import TypeVar, Iterable, Generic, Type, Optional, Dict, Union
from collections.abc import Mapping
from functools import lru_cache
import operator
import pprint

from dataclassy import dataclass

from ..formats import dll


@dataclass(kwargs=True, slots=True)
class Entity:
    """
    The base data class for any entity defined within Freelancer, distinguished by a nickname.

    Attributes (fields) of Entity classes represent entries defined in the INI file section type which this class
    represents. This allows Entities to be automatically constructed from parsed INI files. Fields may have default
    values (defaults are inferred from the input Freelancer expects rather than being defined in the INI files).

    Methods represent derived fields. For example, the ids_name attribute stores the resource ID of an Entity's name as
    defined in the INI. The name() method looks up this resource ID in the resource table and returns the string it
    refers to.

    Attributes not present in the INIs but useful to be passed at construction time /can/ be defined, though this should
    be kept to a minimum. These attributes must have names beginning with _ to denote their internal status.

    If you want to extend these classes to cover non-standard fields (e.g. for an unsupported mod), you can use
    dataclassy.create_dataclass to dynamically define a dataclass and then use these classes as mixins.
    """
    nickname: str  # a unique string identifier for this entity
    ids_name: Optional[int] = None  # resource id for name. newcharacter and universe.ini call this strid_name instead
    ids_info: Optional[int] = None  # resource id for infocard

    def name(self) -> str:
        """The display name of this entity."""
        return dll.lookup(self.ids_name)

    def infocard(self, markup='html') -> str:
        """The infocard for this entity, formatted in the markup language (`rdl`, `html` or `plain`) specified."""
        if markup == 'html':
            lookup = dll.lookup_as_html
        elif markup == 'plain':
            lookup = dll.lookup_as_plain
        elif markup == 'rdl':
            lookup = dll.lookup
        else:
            raise ValueError
        return lookup(self.ids_info)

    def __hash__(self) -> int:
        return hash(self.nickname)

    def __eq__(self, other) -> bool:
        return self.nickname == other.nickname


T = TypeVar('T')
F = TypeVar('F')


class EntitySet(Mapping, Generic[T]):
    """An immutable collection of entities, indexed by nickname."""
    pprint.sorted = lambda v, key=None: v  # patch pprint's sorted implementation to print in insertion order

    def __init__(self, entities: Union[Iterable[T], Dict[str, T]]):
        if type(entities) is dict:
            self._map = entities
        else:
            self._map = {e.nickname: e for e in entities}

    def __repr__(self):
        return f'EntitySet({pprint.pformat(self._map)})'

    def __getitem__(self, key: str) -> T:
        if type(key) is not str:
            raise TypeError(f'Only strings may be used as indices, not {type(key)!r}')
        assert type(key) is str, repr(type(key))
        return self._map[key]

    def __iter__(self):
        """Iteration is over values."""
        return iter(self._map.values())

    def __contains__(self, item):
        """Membership checking is as per hash table."""
        return type(item) is str and item in self._map

    def __len__(self):
        """Length is the size of the map."""
        return len(self._map)

    def __eq__(self, other: 'EntitySet'):
        """EntitySets may be compared by comparing their maps."""
        if not isinstance(other, EntitySet):
            raise TypeError(f'Cannot compare EntitySet with {type(other)!r}')
        return self._map == other._map

    def __hash__(self):
        """The set of keys is constant for an EntitySet and allows it to be hashed."""
        return hash(frozenset(self._map))

    def __add__(self, other) -> 'EntitySet[T]':
        """Two EntitySets can be added together to create a new EntitySet."""
        if type(other) is not type(self):
            raise TypeError(f'Can only concatenate EntitySet (not {type(other)}) with EntitySet.')
        return EntitySet({e for e in self} | {e for e in other})

    def __iadd__(self, other) -> 'EntitySet[T]':
        """An EntitySet can be extended."""
        return self + other

    @lru_cache(maxsize=256)
    def of_type(self, type_: Type[F]) -> 'EntitySet[F]':
        """Return a new, homogeneous EntitySet containing only Entities which are instances of the given type."""
        return EntitySet(filter(lambda e: isinstance(e, type_), self))

    @lru_cache(maxsize=256)
    def reindex(self, on: str) -> 'EntitySet[T]':
        """Reindex this EntitySet on the field with name `on`."""
        return EntitySet({getattr(entity, on): entity for entity in self})

    def where(self, op=operator.eq, **kwargs) -> 'EntitySet[T]':
        """Return a new EntitySet containing only Entities for which the given field matches the given condition.
        Attributes and methods which do not take an argument can be used as fields.
        Usage example: `systems.where(name='New Berlin')`.

        The parameter `op` specifies the comparison operation to be performed on the field. It defaults to testing
        for equality.

        For more complicated queries, use Python's own conditional generator expressions:
        E.g. EntitySet(s for s in systems if s.nickname.startswith('rh'))"""
        assert len(kwargs) == 1
        field, value = next(iter(kwargs.items()))
        if not callable(getattr(self.first, field, None)):
            return self.reindex(field)
        return EntitySet(e for e in self if op(vars(e).get(field) or getattr(e, field)(), value))

    def unique(self, **kwargs) -> Optional[T]:
        """Analogous to `where` but returns the Entity which is guaranteed to uniquely match the specified field/value
        combination. If no such Entity is present, return None."""
        assert len(kwargs) == 1
        field, value = next(iter(kwargs.items()))
        return self.reindex(field).get(value)

    @property
    def first(self) -> Optional[T]:
        """Return the first entity in the set, or None if it is empty. This is useful both for testing. If a query is
        expected to return exactly one result, you likely want to use unique() instead of this property."""
        return next(iter(self), None)


# exported types
from .equipment import *
from .goods import *
from .ship import *
from .solars import *
from .universe import *



"""
Copyright (C) 2016, 2017, 2020 biqqles.

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
from typing import Tuple, Optional, Union

from dataclassy import dataclass

from . import Entity
from .. import maps
from ..formats import dll


class Solar(Entity):
    """A solar is something fixed in space (this name comes from the DATA/SOLAR directory)."""
    ids_info = None
    pos: maps.PosVector  # position vector for this solar
    rotate: maps.RotVector = maps.RotVector(0, 0, 0)  # rotation vector for this solar (defaults to no rotation)
    _system: 'System'  # the system this solar resides in

    def sector(self) -> str:
        """The human-readable navmap coordinate (the centre of) this solar resides in."""
        return maps.pos_to_sector(self.pos, self._system.navmapscale)

    def system(self) -> 'System':
        """The entity of the system this solar resides in."""
        return self._system


class Object(Solar):
    """Generic class for a celestial body - a solid object in space. Objects are automatically classified into
    subclasses in `routines.get_system_contents`."""
    archetype: str


class Jump(Object):
    """A jump conduit is a wormhole - artificial or natural - between star systems."""
    goto: Tuple[str, str, str]

    def type(self):
        """Return a human readable name of this jump conduits's type."""
        if 'gate' in self.archetype: return 'Jump Gate'
        if 'jumphole' in self.archetype: return 'Jump Hole'
        if self.archetype == 'entrypoint': return 'Atmospheric Entry'
        return 'Unknown'

    def origin_system(self) -> 'System':
        """The system this wormhole starts in."""
        return self._system

    def destination_system(self) -> 'System':
        """The system this wormhole ends in."""
        destination = self.goto[-1][0] if isinstance(self.goto, list) else self.goto[0]
        return routines.get_systems()[destination]


class TradeLaneRing(Object):
    """A trade lane ring is a component of a trade lane, a structure which provides "superluminal travel" within a
    system."""
    prev_ring: Optional[str] = None
    next_ring: Optional[str] = None


class Wreck(Object):
    """A wreck (called "secrets" in the game files) is a lootable, wrecked ship floating in space."""
    loadout: str  # loot that is dropped upon being shot


@dataclass(slots=False)
class BaseSolar(Object):
    """The physical representation of a Base."""
    reputation: str  # the nickname of the group this base belongs to
    base: str  # the base (in universe.ini) this solar represents

    def universe_base(self) -> 'Base':
        """The Base entity this solar represents."""
        return routines.get_bases().get(self.base)

    def owner(self) -> 'Faction':
        """The Faction entity that operates this base."""
        return routines.get_factions()[self.reputation]

    def infocard(self, markup='html') -> str:
        """Base infocards are actually in two parts, with ids_info referring to the specs of a base and ids_info + 1
        storing the actual description"""
        if markup == 'html':
            lookup = dll.lookup_as_html
        elif markup == 'plain':
            lookup = dll.lookup_as_plain
        elif markup == 'rdl':
            lookup = dll.lookup
        else:
            raise ValueError

        specifications = lookup(self.ids_info)
        try:
            synopsis = lookup(self.ids_info + 1)
            return specifications + '<p>' + synopsis
        except KeyError:
            return specifications


class Spheroid(Object):
    """A star or planet. (Abstract.)"""
    atmosphere_range: int = 0


class Star(Spheroid):
    """A star in a System."""
    star: str


@dataclass(slots=False)
class Planet(Spheroid):
    """A planet in a System."""
    spin: Tuple[float, float, float]


class PlanetaryBase(Planet, BaseSolar):
    """A base on the surface of a planet, typically accessible via a docking ring."""


class Zone(Solar):
    """A zone is a region of space, possibly with effects attached."""
    size: Union[int, Tuple[int, int], Tuple[int, int, int]]
    shape: str  # one of: sphere, ring, box, ellipsoid


from .universe import Base, Faction, System
from .. import routines
